// Copyright 2014 Google Inc. All rights reserved.
//
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file or at
// https://developers.google.com/open-source/licenses/bsd

#ifndef __U2F_H_INCLUDED__
#define __U2F_H_INCLUDED__

#include <stdint.h>

#ifdef __cplusplus
extern "C" {
#endif

#ifndef __NO_PRAGMA_PACK
#pragma pack(push, 1)
#endif

// ISO 7816 status codes
#define SW_NO_ERROR                 0x9000
#define SW_CONDITIONS_NOT_SATISFIED 0x6985
#define SW_WRONG_INSTRUCTION        0x6D00
#define SW_WRONG_LENGTH             0x6700
#define SW_CLASS_NOT_SUPPORTED      0x6E00
#define SW_WRONG_DATA_PARAMETERS    0x6A80
#define SW_WRONG_DATA		    0x6984
// General constants

#define P256_SCALAR_SIZE  32  // nistp256 in bytes
#define P256_POINT_SIZE  ((P256_SCALAR_SIZE * 2) + 1)

//#define MAX_ECDSA_SIG_SIZE  72  // asn1 DER format
//#define MAX_KH_SIZE  128  // key handle
//#define MAX_CERT_SIZE  2048  // attestation certificate

#define U2F_EC_KEY_SIZE 		32
#define U2F_EC_POINT_SIZE 		((U2F_EC_KEY_SIZE * 2) + 1) // Size of EC point
#define U2F_MAX_KH_SIZE 		128
#define U2F_MAX_ATT_CERT_SIZE 		2048
#define U2F_MAX_EC_SIG_SIZE 		72
#define U2F_CTR_SIZE			4
#define U2F_APPID_SIZE 			32
#define U2F_CHAL_SIZE			32


//#define U2F_APPID_SIZE  32
#define U2F_NONCE_SIZE  32

#define UNCOMPRESSED_POINT 0x04

typedef struct {
  uint8_t format;
  uint8_t x[U2F_EC_KEY_SIZE];
  uint8_t y[U2F_EC_KEY_SIZE];
} P256_POINT;

// U2Fv2 instructions

#define U2F_INS_REGISTER  	0x01
#define U2F_INS_AUTHENTICATE  	0x02
#define U2F_INS_VERSION  	0x03

// U2F_REGISTER instruction defines

#define U2F_REGISTER_ID  	0x05  // magic constant
#define U2F_REGISTER_HASH_ID  	0x00  // magic constant

typedef struct {
  uint8_t nonce[U2F_CHAL_SIZE];
  uint8_t appId[U2F_APPID_SIZE];
} U2F_REGISTER_REQ;

typedef struct {
  uint8_t registerId;
  P256_POINT pubKey;
  uint8_t keyHandleLen;
  uint8_t keyHandleCertSig[
      U2F_MAX_KH_SIZE +
      U2F_MAX_ATT_CERT_SIZE +
      U2F_MAX_EC_SIG_SIZE];
} U2F_REGISTER_RESP;

// U2F_AUTHENTICATE instruction defines

// Authentication parameter byte
#define U2F_AUTH_ENFORCE        0x03  // Require user presence
#define U2F_AUTH_CHECK_ONLY     0x07  // Test but do not consume
#define U2F_AUTH_FLAG_TUP       0x01  // Test user presence

typedef struct {
  uint8_t nonce[U2F_CHAL_SIZE];
  uint8_t appId[U2F_APPID_SIZE];
  uint8_t keyHandleLen;
  uint8_t keyHandle[U2F_MAX_KH_SIZE];
} U2F_AUTHENTICATE_REQ;

// Flags values
#define U2F_TOUCHED  0x01
#define U2F_ALTERNATE_INTERFACE  0x02

typedef struct {
  uint8_t flags;
  uint32_t ctr;
  uint8_t sig[U2F_MAX_EC_SIG_SIZE];
} U2F_AUTHENTICATE_RESP;

#ifndef __NO_PRAGMA_PACK
#pragma pack(pop)
#endif

#ifdef __cplusplus
}
#endif

#endif  // __U2F_H_INCLUDED__
